"""Maps client objects from NSB Sharepoint database to internal AIND models."""

from typing import List

from aind_data_schema.procedures import SubjectProcedure
from office365.sharepoint.client_context import ClientContext

from aind_metadata_service.sharepoint.nsb2019.mapping import MappedNSBList
from aind_metadata_service.sharepoint.nsb2019.models import NSBList


class NSB2019Procedures:
    """Provides methods to retrieve procedure information from sharepoint,
    parses the response into an intermediate data model, and maps that model
    into AIND Procedures model."""

    _view_title = "New Request"

    def get_procedures_from_sharepoint(
        self, subject_id: str, client_context: ClientContext, list_title: str
    ) -> List[SubjectProcedure]:
        """
        Get list of Procedures from NSB 2019 database.
        Parameters
        ----------
        subject_id : str
          ID of the subject to find procedure information for
        client_context : ClientContext
          NSB Sharepoint client
        list_title : str
          Title of the list where the 2019 procedure data is stored
        Returns
        -------
        List[SubjectProcedure]
        """

        labtrack_alias = NSBList.__fields__.get("lab_tracks_id").alias
        filter_string = f"{labtrack_alias} eq '{subject_id}'"
        list_view = client_context.web.lists.get_by_title(
            list_title
        ).views.get_by_title(self._view_title)
        client_context.load(list_view)
        client_context.execute_query()
        list_items = list_view.get_items().filter(filter_string)
        client_context.load(list_items)
        client_context.execute_query()
        list_of_procedures = []
        for list_item in list_items:
            nsb_model = NSBList.parse_obj(list_item.to_json())
            mapped_model = MappedNSBList(nsb=nsb_model)
            procedures = mapped_model.get_procedures()
            list_of_procedures.extend(procedures)
        return list_of_procedures



""""Tests response_handler module"""

import json
import os
import unittest
from pathlib import Path

from aind_data_schema.procedures import Procedures
from aind_data_schema.subject import Subject
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from pydantic import validate_model

from aind_metadata_service.response_handler import Responses

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
DIR_MAP = TEST_DIR / "resources" / "json_responses"

SP_RESPONSE_PATH = DIR_MAP / "mapped_sp_procedure.json"
LAS_RESPONSE_PATH = DIR_MAP / "mapped_las_procedure.json"
COMBINED_PATH = DIR_MAP / "combined.json"

with open(SP_RESPONSE_PATH) as f:
    sp_subject_procedures = json.load(f)

with open(LAS_RESPONSE_PATH) as f:
    las_subject_procedures = json.load(f)

with open(COMBINED_PATH) as f:
    combined_procedures = json.load(f)


class TestResponseHandler(unittest.TestCase):
    """Tests methods in Responses class"""

    def test_valid_model(self):
        """Test model_response with valid model."""

        model = Subject(
            species="Mus musculus",
            subject_id="639374",
            sex="Male",
            date_of_birth="2022-06-24",
            genotype="Pvalb-IRES-Cre/wt;RCL-somBiPoles_mCerulean-WPRE/wt",
        )
        response = Responses.model_response(model)
        model_json = jsonable_encoder(model)
        expected_response = JSONResponse(
            status_code=200,
            content=(
                {
                    "message": "Valid Model.",
                    "data": model_json,
                }
            ),
        )

        self.assertEqual(expected_response.status_code, response.status_code)
        self.assertEqual(expected_response.body, response.body)

    def test_invalid_model(self):
        """Test model_response with invalid model."""
        model = Subject.construct()
        response = Responses.model_response(model)
        *_, validation_error = validate_model(model.__class__, model.__dict__)
        model_json = jsonable_encoder(model)
        expected_response = JSONResponse(
            status_code=406,
            content=(
                {
                    "message": f"Validation Errors: {validation_error}",
                    "data": model_json,
                }
            ),
        )

        self.assertEqual(expected_response.status_code, response.status_code)
        self.assertEqual(expected_response.body, response.body)

    def test_connection_error(self):
        """Test connection error response"""
        response = Responses.connection_error_response()
        expected_response = JSONResponse(
            status_code=503,
            content=(
                {
                    "message": "Error Connecting to Internal Server.",
                    "data": None,
                }
            ),
        )
        self.assertEqual(expected_response.status_code, response.status_code)
        self.assertEqual(expected_response.body, response.body)

    def test_combine_internal_error_responses(self):
        """Tests that error responses are combined as expected"""
        response1 = Responses.internal_server_error_response()
        response2 = Responses.internal_server_error_response()
        response = Responses.combine_procedure_responses(
            lb_response=response1, sp_response=response2
        )
        expected_response = Responses.internal_server_error_response()
        self.assertEqual(response.body, expected_response.body)
        self.assertEqual(response.status_code, 500)

    def test_combine_connection_error_responses(self):
        """Tests that error responses are combined as expected"""
        response1 = Responses.connection_error_response()
        response2 = Responses.connection_error_response()
        response = Responses.combine_procedure_responses(
            lb_response=response1, sp_response=response2
        )
        expected_response = Responses.connection_error_response()
        self.assertEqual(response.body, expected_response.body)
        self.assertEqual(response.status_code, 503)

    def test_combine_no_data_error_responses(self):
        """Tests that error responses are combined as expected"""
        response1 = Responses.no_data_found_response()
        response2 = Responses.no_data_found_response()
        response = Responses.combine_procedure_responses(
            lb_response=response1, sp_response=response2
        )
        expected_response = Responses.no_data_found_response()
        self.assertEqual(response.body, expected_response.body)
        self.assertEqual(response.status_code, 404)

    def test_combine_valid_no_data_error_responses(self):
        """Tests that error responses are combined as expected"""
        response1 = Responses.no_data_found_response()
        response2 = JSONResponse(las_subject_procedures)
        response3 = JSONResponse(sp_subject_procedures)
        combined_response_1 = Responses.combine_procedure_responses(
            lb_response=response2, sp_response=response1
        )
        expected_response_1 = JSONResponse(las_subject_procedures)
        self.assertEqual(combined_response_1.body, expected_response_1.body)
        self.assertEqual(combined_response_1.status_code, 200)

        combined_response_2 = Responses.combine_procedure_responses(
            lb_response=response1, sp_response=response3
        )
        expected_response_2 = JSONResponse(sp_subject_procedures)
        self.assertEqual(combined_response_2.body, expected_response_2.body)
        self.assertEqual(combined_response_2.status_code, 200)

    def test_combine_valid_responses(self):
        """Tests that valid responses are combined as expected"""
        response1 = JSONResponse(las_subject_procedures)
        response2 = JSONResponse(sp_subject_procedures)
        response = Responses.combine_procedure_responses(
            lb_response=response1, sp_response=response2
        )
        expected_response = JSONResponse(combined_procedures)
        self.assertEqual(expected_response.body, response.body)
        self.assertEqual(200, response.status_code)

    def test_combine_invalid_responses(self):
        """Tests that invalid responses are combined as expected"""
        model1 = Procedures.construct(
            subject_id="115977",
            extra_field=None,
            subject_procedures=sp_subject_procedures["data"][
                "subject_procedures"
            ],
        )
        *_, validation_error_1 = validate_model(
            model1.__class__, model1.__dict__
        )
        model2 = Procedures.construct(
            subject_id="115977",
            subject_procedures=las_subject_procedures["data"][
                "subject_procedures"
            ],
        )
        *_, validation_error_2 = validate_model(
            model2.__class__, model2.__dict__
        )
        response1 = Responses.model_response(model1)
        response2 = Responses.model_response(model2)
        response = Responses.combine_procedure_responses(
            lb_response=response2, sp_response=response1
        )
        expected_response = JSONResponse(
            status_code=406,
            content=(
                {
                    "message": f"Message 1: "
                    f"Validation Errors: {str(validation_error_1)}, "
                    f"Message 2: "
                    f"Validation Errors: {str(validation_error_2)}",
                    "data": combined_procedures["data"],
                }
            ),
        )
        self.assertEqual(response.body, expected_response.body)
        self.assertEqual(response.status_code, expected_response.status_code)

    def test_combine_valid_invalid_responses(self):
        """Tests that valid and invalid response are combined as expected"""
        model = Procedures.construct()
        response1 = JSONResponse(sp_subject_procedures)
        response2 = Responses.model_response(model)
        *_, validation_error = validate_model(model.__class__, model.__dict__)
        response = Responses.combine_procedure_responses(
            lb_response=response2, sp_response=response1
        )
        expected_response = JSONResponse(
            status_code=207,
            content=(
                {
                    "message": f"Message 1: Valid Model., "
                    f"Message 2: Validation Errors: {str(validation_error)}",
                    "data": sp_subject_procedures["data"],
                }
            ),
        )
        self.assertEqual(response.body, expected_response.body)
        self.assertEqual(response.status_code, expected_response.status_code)

    def test_combine_valid_error_responses(self):
        """Tests that error responses are combined as expected"""
        response1 = Responses.connection_error_response()
        response2 = JSONResponse(sp_subject_procedures)
        combined_response_1 = Responses.combine_procedure_responses(
            lb_response=response1, sp_response=response2
        )
        expected_response_1 = JSONResponse(
            status_code=207,
            content=(
                {
                    "message": "Message 1: Valid Model., "
                    "Message 2: Error Connecting to Internal Server.",
                    "data": sp_subject_procedures["data"],
                }
            ),
        )
        self.assertEqual(combined_response_1.body, expected_response_1.body)
        self.assertEqual(
            combined_response_1.status_code, expected_response_1.status_code
        )

        response3 = JSONResponse(las_subject_procedures)
        combined_response_2 = Responses.combine_procedure_responses(
            lb_response=response3, sp_response=response1
        )
        expected_response_2 = JSONResponse(
            status_code=207,
            content=(
                {
                    "message": "Message 1: "
                    "Error Connecting to Internal Server., "
                    "Message 2: Valid Model.",
                    "data": las_subject_procedures["data"],
                }
            ),
        )
        self.assertEqual(combined_response_2.body, expected_response_2.body)
        self.assertEqual(
            combined_response_2.status_code, expected_response_2.status_code
        )


if __name__ == "__main__":
    unittest.main()



"""Tests NSB 2023 data models are created correctly"""

import json
import logging
import os
from pathlib import Path
from typing import List
from unittest import TestCase
from unittest import main as unittest_main

from aind_metadata_service.sharepoint.nsb2023.models import NSBList

if os.getenv("LOG_LEVEL"):  # pragma: no cover
    logging.basicConfig(level=os.getenv("LOG_LEVEL"))

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / ".." / ".."
SHAREPOINT_DIR = TEST_DIR / "resources" / "sharepoint" / "nsb2023" / "raw"
LIST_ITEM_FILE_NAMES = os.listdir(SHAREPOINT_DIR)
LIST_ITEM_FILE_PATHS = [SHAREPOINT_DIR / str(f) for f in LIST_ITEM_FILE_NAMES]
LIST_ITEM_FILE_PATHS.sort()


class TestNSB2023Models(TestCase):
    """Tests methods in NSBList2023 class"""

    @classmethod
    def setUpClass(cls):
        """Load json files before running tests."""
        cls.list_items = cls._load_json_files()

    @staticmethod
    def _load_json_files() -> List[dict]:
        """Reads test data into json"""
        list_items = []
        for file_path in LIST_ITEM_FILE_PATHS:
            with open(file_path) as f:
                contents = json.load(f)
            list_items.append(contents)
        return list_items

    def test_models_parsed(self):
        """Tests that the given list item files are parsed without error."""
        for index in range(len(self.list_items)):
            list_item = self.list_items[index]
            logging.debug(f"Processing file: {LIST_ITEM_FILE_NAMES[index]}")
            nsb_model = NSBList.parse_obj(list_item)
            self.assertEqual(
                nsb_model.author_id, int(list_item.get("AuthorId"))
            )


if __name__ == "__main__":
    unittest_main()



"""Module to handle responses"""
import json
from typing import List

from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from pydantic import BaseModel, validate_model

from aind_metadata_service.client import StatusCodes


class Responses:
    """This class contains methods to map responses from server."""

    @staticmethod
    def multi_status_response(
        message1: str, message2: str, data
    ) -> JSONResponse:
        """Map to a multi status message."""
        response = JSONResponse(
            status_code=StatusCodes.MULTI_STATUS.value,
            content=(
                {
                    "message": f"Message 1: {message1}, "
                    f"Message 2: {message2}",
                    "data": data,
                }
            ),
        )
        return response

    @staticmethod
    def connection_error_response() -> JSONResponse:
        """Map to a connection error"""
        response = JSONResponse(
            status_code=StatusCodes.CONNECTION_ERROR.value,
            content=(
                {
                    "message": "Error Connecting to Internal Server.",
                    "data": None,
                }
            ),
        )
        return response

    @staticmethod
    def internal_server_error_response() -> JSONResponse:
        """Map to an internal server error"""
        response = JSONResponse(
            status_code=StatusCodes.INTERNAL_SERVER_ERROR.value,
            content=({"message": "Internal Server Error.", "data": None}),
        )
        return response

    @staticmethod
    def no_data_found_response() -> JSONResponse:
        """Map to a 404 error."""
        response = JSONResponse(
            status_code=StatusCodes.NO_DATA_FOUND.value,
            content=({"message": "No Data Found.", "data": None}),
        )
        return response

    @staticmethod
    def multiple_items_found_response(models: List[BaseModel]) -> JSONResponse:
        """Map to a multiple choices error."""
        models_json = [jsonable_encoder(model) for model in models]
        response = JSONResponse(
            status_code=StatusCodes.MULTIPLE_RESPONSES.value,
            content=(
                {"message": "Multiple Items Found.", "data": models_json}
            ),
        )
        return response

    @staticmethod
    def model_response(model: BaseModel) -> JSONResponse:
        """
        Parse model to a response or return model if valid.
        Handles validation errors.
        Parameters
        ----------
        model : BaseModel
            model response from server
        """
        *_, validation_error = validate_model(model.__class__, model.__dict__)
        model_json = jsonable_encoder(model)
        if validation_error:
            response = JSONResponse(
                status_code=StatusCodes.INVALID_DATA.value,
                content=(
                    {
                        "message": f"Validation Errors: {validation_error}",
                        "data": model_json,
                    }
                ),
            )
        else:
            response = JSONResponse(
                status_code=StatusCodes.VALID_DATA.value,
                content=(
                    {
                        "message": "Valid Model.",
                        "data": model_json,
                    }
                ),
            )
        return response

    # flake8: noqa: C901
    @staticmethod
    def combine_procedure_responses(
        lb_response: JSONResponse, sp_response: JSONResponse
    ) -> JSONResponse:
        """
        Combines JSONResponses from Labtracks and Sharepoint clients.
        Handles validation errors and special cases.
        """
        lb_contents = json.loads(lb_response.body)
        sp_contents = json.loads(sp_response.body)
        sp_status_code = sp_response.status_code
        lb_status_code = lb_response.status_code
        if sp_status_code == 500 and lb_status_code == 500:
            return Responses.internal_server_error_response()
        if sp_status_code == 503 and lb_status_code == 503:
            return Responses.connection_error_response()
        if sp_status_code == 404 and lb_status_code == 404:
            return Responses.no_data_found_response()
        if sp_status_code < 300 and lb_status_code == 404:
            return sp_response
        if sp_status_code == 404 and lb_status_code < 300:
            return lb_response
        # handles combination of invalid responses
        if lb_status_code == 406 and sp_status_code == 406:
            lb_data = lb_contents["data"]["subject_procedures"]
            sp_data = sp_contents["data"]["subject_procedures"]
            combined_data = lb_data + sp_data
            lb_contents["data"]["subject_procedures"] = combined_data
            message1 = sp_contents["message"]
            message2 = lb_contents["message"]
            return JSONResponse(
                status_code=StatusCodes.INVALID_DATA.value,
                content=(
                    {
                        "message": f"Message 1: {message1}, "
                        f"Message 2: {message2}",
                        "data": lb_contents["data"],
                    }
                ),
            )
        # handles combinations of valid and invalid responses
        if (sp_status_code < 300 and lb_status_code == 406) or (
            lb_status_code < 300 and sp_status_code == 406
        ):
            lb_data = lb_contents["data"]["subject_procedures"]
            sp_data = sp_contents["data"]["subject_procedures"]
            combined_data = lb_data + sp_data
            sp_contents["data"]["subject_procedures"] = combined_data
            return Responses.multi_status_response(
                message1=sp_contents["message"],
                message2=lb_contents["message"],
                data=sp_contents["data"],
            )
        # handles case when both responses are valid
        if sp_status_code < 300 and lb_status_code < 300:
            lb_data = lb_contents["data"]["subject_procedures"]
            sp_data = sp_contents["data"]["subject_procedures"]
            combined_data = lb_data + sp_data
            lb_contents["data"]["subject_procedures"] = combined_data
            return JSONResponse(
                status_code=StatusCodes.VALID_DATA.value,
                content=(
                    {"message": "Valid Model.", "data": lb_contents["data"]}
                ),
            )
        # handles combination of server/connection error and valid response
        if sp_status_code in (500, 503) and lb_status_code < 300:
            lb_data = lb_contents["data"]["subject_procedures"]
            lb_contents["data"]["subject_procedures"] = lb_data
            return Responses.multi_status_response(
                message1=sp_contents["message"],
                message2=lb_contents["message"],
                data=lb_contents["data"],
            )
        if sp_status_code < 300 and lb_status_code in (500, 503):
            sp_data = sp_contents["data"]["subject_procedures"]
            sp_contents["data"]["subject_procedures"] = sp_data
            return Responses.multi_status_response(
                message1=sp_contents["message"],
                message2=lb_contents["message"],
                data=sp_contents["data"],
            )



"""Starts and runs a Flask Service"""

import os

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse

from aind_metadata_service.labtracks.client import LabTracksClient
from aind_metadata_service.response_handler import Responses
from aind_metadata_service.sharepoint.client import SharePointClient

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["GET"],
    allow_headers=["*"],
)

# TODO: Handle configs better?
favicon_path = os.getenv("FAVICON_PATH")

labtracks_driver = os.getenv("ODBC_DRIVER")
labtracks_server = os.getenv("LABTRACKS_SERVER")
labtracks_port = os.getenv("LABTRACKS_PORT")
labtracks_db = os.getenv("LABTRACKS_DATABASE")
labtracks_user = os.getenv("LABTRACKS_USER")
labtracks_password = os.getenv("LABTRACKS_PASSWORD")

nsb_sharepoint_url = os.getenv("NSB_SHAREPOINT_URL")
nsb_sharepoint_list_2019 = os.getenv("NSB_2019_LIST", default="SWR 2019-2022")
nsb_sharepoint_list_2023 = os.getenv(
    "NSB_2023_LIST", default="SWR 2023-Present"
)
nsb_sharepoint_user = os.getenv("NSB_SHAREPOINT_USER")
nsb_sharepoint_password = os.getenv("NSB_SHAREPOINT_PASSWORD")


@app.get("/subject/{subject_id}")
async def retrieve_subject(subject_id):
    """
    Retrieves subject from Labtracks server
    """
    lb_client = LabTracksClient(
        driver=labtracks_driver,
        server=labtracks_server,
        port=labtracks_port,
        db=labtracks_db,
        user=labtracks_user,
        password=labtracks_password,
    )
    response = lb_client.get_subject_info(subject_id)
    return response


@app.get("/procedures/{subject_id}")
async def retrieve_procedures(subject_id):
    """
    Retrieves procedure info from SharePoint
    """
    sharepoint_client = SharePointClient(
        nsb_site_url=nsb_sharepoint_url,
        nsb_list_title_2019=nsb_sharepoint_list_2019,
        nsb_list_title_2023=nsb_sharepoint_list_2023,
        client_id=nsb_sharepoint_user,
        client_secret=nsb_sharepoint_password,
    )
    lb_client = LabTracksClient(
        driver=labtracks_driver,
        server=labtracks_server,
        port=labtracks_port,
        db=labtracks_db,
        user=labtracks_user,
        password=labtracks_password,
    )
    lb_response = lb_client.get_procedures_info(subject_id=subject_id)
    sp_response = sharepoint_client.get_procedure_info(subject_id=subject_id)
    response = Responses.combine_procedure_responses(lb_response, sp_response)
    return response


@app.get(
    "/favicon.ico",
    include_in_schema=False,
    response_class=RedirectResponse,
    status_code=200,
)
async def favicon():
    """
    Returns the favicon
    """
    return favicon_path



"""Module to test the aind_metadata_service client."""
import json
import unittest
from unittest import mock
from unittest.mock import MagicMock, call

import requests

from aind_metadata_service.client import AindMetadataServiceClient


class TestAindMetadataServiceClient(unittest.TestCase):
    """Tests the AindMetadataServiceClient class methods"""

    @mock.patch("requests.get")
    def test_get_subject(self, mock_get: MagicMock) -> None:
        """Tests the get_subject method."""

        mock_subject_id = "00000"
        mock_response = requests.Response()
        mock_response.status_code = 200
        mock_response._content = json.dumps(
            {
                "message": "Valid Model.",
                "data": {
                    "describedBy": "https://acme.com/subject.py",
                    "schema_version": "0.2.2",
                    "species": "Mus musculus",
                    "subject_id": mock_subject_id,
                    "sex": "Female",
                    "date_of_birth": "2022-05-01",
                    "genotype": "wt/wt",
                    "mgi_allele_ids": None,
                    "background_strain": None,
                    "source": None,
                    "rrid": None,
                    "restrictions": None,
                    "breeding_group": "breeding_group_id",
                    "maternal_id": "00001",
                    "maternal_genotype": "wt/wt",
                    "paternal_id": "00002",
                    "paternal_genotype": "wt/wt",
                    "light_cycle": None,
                    "home_cage_enrichment": None,
                    "wellness_reports": None,
                    "notes": None,
                },
            }
        ).encode("utf-8")

        mock_get.return_value.__enter__.return_value = mock_response

        ams_client = AindMetadataServiceClient(domain="some_url")

        response = ams_client.get_subject(mock_subject_id)

        mock_get.assert_has_calls(
            [
                call("some_url/subject/00000"),
                call().__enter__(),
                call().__exit__(None, None, None),
            ]
        )

        self.assertEqual(200, response.status_code)
        self.assertEqual(mock_response.json(), response.json())

    @mock.patch("requests.get")
    def test_get_procedures(self, mock_get: MagicMock) -> None:
        """Tests the get_procedures method."""

        mock_subject_id = "00000"
        mock_response = requests.Response()
        mock_response.status_code = 418
        mock_response._content = json.dumps(
            {
                "message": "Validation Errors: 8 validation errors for "
                "Procedures\nheadframes\n  'NoneType' object is not "
                "iterable (type=type_error)\ncraniotomies\n  "
                "'NoneType' object is not iterable "
                "(type=type_error)\nmri_scans\n  'NoneType' object is "
                "not iterable (type=type_error)\ninjections\n  "
                "'NoneType' object is not iterable "
                "(type=type_error)\nfiber_implants\n  "
                "'NoneType' object is not iterable "
                "(type=type_error)\ntraining_protocols\n  "
                "'NoneType' object is not iterable "
                "(type=type_error)\ntissue_preparations\n  "
                "'NoneType' object is not iterable "
                "(type=type_error)\nother_procedures\n  "
                "'NoneType' object is not iterable (type=type_error)",
                "data": {
                    "describedBy": "https://acme.com/procedures.py",
                    "schema_version": "0.4.4",
                    "subject_id": mock_subject_id,
                    "headframes": None,
                    "craniotomies": None,
                    "mri_scans": None,
                    "injections": None,
                    "fiber_implants": None,
                    "water_restriction": None,
                    "training_protocols": None,
                    "tissue_preparations": None,
                    "other_procedures": None,
                    "notes": None,
                },
            }
        ).encode("utf-8")

        mock_get.return_value.__enter__.return_value = mock_response

        ams_client = AindMetadataServiceClient(domain="some_url")

        response = ams_client.get_procedures(mock_subject_id)

        mock_get.assert_has_calls(
            [
                call("some_url/procedures/00000"),
                call().__enter__(),
                call().__exit__(None, None, None),
            ]
        )

        self.assertEqual(418, response.status_code)
        self.assertEqual(mock_response.json(), response.json())


if __name__ == "__main__":
    unittest.main()



"""Module to create client to connect to sharepoint database"""

from aind_data_schema.procedures import Procedures
from fastapi.responses import JSONResponse
from office365.runtime.auth.client_credential import ClientCredential
from office365.sharepoint.client_context import ClientContext

from aind_metadata_service.response_handler import Responses
from aind_metadata_service.sharepoint.nsb2019.procedures import (
    NSB2019Procedures,
)
from aind_metadata_service.sharepoint.nsb2023.procedures import (
    NSB2023Procedures,
)


class SharePointClient:
    """This class contains the api to connect to SharePoint db."""

    def __init__(
        self,
        nsb_site_url: str,
        nsb_list_title_2019: str,
        nsb_list_title_2023: str,
        client_id: str,
        client_secret: str,
    ) -> None:
        """
        Initialize a client
        Parameters
        ----------
        nsb_site_url : str
           sharepoint site url
        client_id : str
            username for principal account to access sharepoint
        client_secret : str
            password for principal account to access sharepoint
        """
        self.nsb_site_url = nsb_site_url
        self.nsb_list_title_2019 = nsb_list_title_2019
        self.nsb_list_title_2023 = nsb_list_title_2023
        self.client_id = client_id
        self.client_secret = client_secret
        self.credentials = ClientCredential(self.client_id, self.client_secret)
        self.nsb_client_context = ClientContext(
            self.nsb_site_url
        ).with_credentials(self.credentials)

    def get_procedure_info(
        self,
        subject_id: str,
    ) -> JSONResponse:
        """
        Primary interface. Maps a subject_id to a response.
        Parameters
        ----------
        subject_id : str
          ID of the subject being queried for.

        Returns
        -------
        JSONResponse
          A response

        """
        # TODO: Add try to handle internal server error response.
        subject_procedures = []
        nsb_ctx = self.nsb_client_context
        nsb_2019_mapper = NSB2019Procedures()
        nsb_2023_mapper = NSB2023Procedures()
        procedures2019 = nsb_2019_mapper.get_procedures_from_sharepoint(
            subject_id=subject_id,
            client_context=nsb_ctx,
            list_title=self.nsb_list_title_2019,
        )
        procedures2023 = nsb_2023_mapper.get_procedures_from_sharepoint(
            subject_id=subject_id,
            client_context=nsb_ctx,
            list_title=self.nsb_list_title_2023,
        )
        subject_procedures.extend(procedures2019)
        subject_procedures.extend(procedures2023)
        response = self._handle_response_from_sharepoint(
            subject_id=subject_id, subject_procedures=subject_procedures
        )
        return response

    @staticmethod
    def _handle_response_from_sharepoint(
        subject_id: str, subject_procedures=None
    ) -> JSONResponse:
        """
        Maps the response from SharePoint into a Procedures model
        Parameters
        ----------
        subject_id : str
          ID of the subject being queried for.
        subject_procedures: None/list

        Returns
        -------
        JSONResponse
          Either a Procedures model or an error response

        """
        if subject_procedures:
            procedures = Procedures.construct(subject_id=subject_id)
            procedures.subject_procedures = subject_procedures
            response = Responses.model_response(procedures)
        else:
            response = Responses.no_data_found_response()
        return response



"""Module to test LabTracksResponseHandler methods."""
import datetime
import unittest
from decimal import Decimal

from aind_data_schema import Procedures, Subject
from aind_data_schema.procedures import Perfusion, RetroOrbitalInjection
from aind_data_schema.subject import BackgroundStrain, Sex, Species

from aind_metadata_service.labtracks.client import (
    LabTracksBgStrain,
    LabTracksResponseHandler,
    LabTracksSex,
    LabTracksSpecies,
)


class TestResponseExamples:
    """Class to hold some examples"""

    class_values_str = (
        '<?xml version="1.0" encoding="utf-16"?>\r\n<MouseCustomClass '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xmlns:xsd="http://www.w3.org/2001/XMLSchema">\r\n '
        "<Reserved_by>Jane Doe</Reserved_by>\r\n "
        "<Reserve_Date>2013-07-10T00:00:00-07:00</Reserve_Date>\r\n "
        "<Reason>Maintenance Breeding</Reason>\r\n "
        "<Full_Genotype>Adora2a-Cre/wt</Full_Genotype>\r\n</MouseCustomClass> "
    )
    paternal_class_values_str = (
        '<?xml version="1.0" encoding="utf-16"?>\r\n<MouseCustomClass '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xmlns:xsd="http://www.w3.org/2001/XMLSchema">\r\n '
        "<Reserved_by>Jane Doe</Reserved_by>\r\n "
        "<Reserve_Date>2013-01-30T00:00:00-08:00</Reserve_Date>\r\n "
        "<Reason>Breeder Refresh</Reason>\r\n "
        "<Full_Genotype>Adora2a-Cre/wt</Full_Genotype>\r\n</MouseCustomClass>"
    )
    maternal_class_values_str = (
        '<?xml version="1.0" encoding="utf-16"?>\r\n<MouseCustomClass '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xmlns:xsd="http://www.w3.org/2001/XMLSchema">\r\n '
        "<Reserved_by>Jane Doe</Reserved_by>\r\n "
        "<Reserve_Date>2013-01-30T00:00:00-08:00</Reserve_Date>\r\n "
        "<Reason>Breeder Refresh</Reason>\r\n "
        "<Full_Genotype>wt/wt</Full_Genotype>\r\n</MouseCustomClass> "
    )
    test_subject_response = [
        {
            "id": Decimal("115977"),
            "class_values": class_values_str,
            "sex": "M",
            "birth_date": datetime.datetime(2012, 5, 13, 0, 0),
            "mouse_comment": None,
            "paternal_id": Decimal("107384"),
            "paternal_class_values": paternal_class_values_str,
            "maternal_id": Decimal("107392"),
            "maternal_class_values": maternal_class_values_str,
            "species_name": "mouse",
            "group_name": "C57BL6J_OLD",
            "group_description": "C57BL/6J",
        }
    ]

    expected_subject = Subject.parse_obj(
        {
            "schema_version": "0.3.0",
            "species": "Mus musculus",
            "subject_id": "115977",
            "sex": "Male",
            "date_of_birth": "2012-05-13",
            "genotype": "Adora2a-Cre/wt",
            "maternal_id": "107392",
            "maternal_genotype": "wt/wt",
            "paternal_id": "107384",
            "paternal_genotype": "Adora2a-Cre/wt",
            "breeding_group": "C57BL6J_OLD",
            "background_strain": "C57BL/6J",
        }
    )

    test_procedures_response = [
        {
            "id": Decimal(00000),
            "task_type_id": Decimal(12345),
            "type_name": "Perfusion Gel",
            "date_start": datetime.datetime(2022, 10, 11, 0, 0),
            "date_end": datetime.datetime(2022, 10, 11, 4, 30),
            "investigator_id": Decimal(28803),
            "task_object": Decimal(115977),
            "protocol_number": Decimal(2002),
        },
        {
            "id": Decimal(10000),
            "task_type_id": Decimal(23),
            "type_name": "RO Injection VGT",
            "date_start": datetime.datetime(2022, 5, 11, 0, 0),
            "date_end": datetime.datetime(2022, 5, 12, 0, 0),
            "investigator_id": Decimal(28803),
            "task_object": Decimal(115977),
            "protocol_number": Decimal(2002),
        },
    ]

    expected_subject_procedures = [
        Perfusion.construct(
            start_date=datetime.date(2022, 10, 11),
            end_date=datetime.date(2022, 10, 11),
            experimenter_full_name=Decimal("28803"),
            iacuc_protocol=Decimal("2002"),
            animal_weight_prior=None,
            animal_weight_post=None,
            anaesthesia=None,
            notes=None,
            procedure_type="Perfusion",
            output_specimen_ids=[Decimal("115977")],
        ),
        RetroOrbitalInjection.construct(
            start_date=datetime.date(2022, 5, 11),
            end_date=datetime.date(2022, 5, 12),
            experimenter_full_name=Decimal("28803"),
            iacuc_protocol=Decimal("2002"),
            animal_weight_prior=None,
            animal_weight_post=None,
            anaesthesia=None,
            notes=None,
            procedure_type="Retro-orbital injection",
        ),
    ]

    expected_procedures = Procedures.parse_obj(
        {
            "schema_version": "0.7.0",
            "subject_id": "115977",
            "subject_procedures": expected_subject_procedures,
        }
    )


class TestLabTracksResponseHandler(unittest.TestCase):
    """Class for unit tests on LabTracksResponseHandler."""

    rh = LabTracksResponseHandler()

    def test_map_response_to_procedures(self):
        """Tests that the response gets mapped to the subject."""
        actual_procedures = self.rh.map_response_to_procedures(
            TestResponseExamples.test_procedures_response
        )
        self.assertEqual(
            TestResponseExamples.expected_subject_procedures, actual_procedures
        )

    def test_map_response_to_subject(self):
        """Tests that the response gets mapped to the subject."""
        actual_subject = self.rh.map_response_to_subject(
            TestResponseExamples.test_subject_response
        )
        self.assertEqual(
            [TestResponseExamples.expected_subject], actual_subject
        )

    def test_map_class_values_to_genotype(self):
        """Tests that the genotype is extracted from the xml string."""
        deformed_class_values_str = (
            '<?xml version="1.0" encoding="utf-16"?>\r\n<MouseCustomClass '
            'xmlns:xsd="http://www.w3.org/2001/XMLSchema" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\r\n  '
            "<Reserved_by>Mary Smith </Reserved_by>\r\n  "
            "<Reserve_Date>2022-06-27T00:00:00</Reserve_Date>\r\n  "
            "<Reason>EU-retire</Reason>\r\n</MouseCustomClass>"
        )

        output = self.rh._map_class_values_to_genotype(
            deformed_class_values_str
        )
        self.assertIsNone(output)
        self.assertIsNone(self.rh._map_class_values_to_genotype(None))

    def test_map_species(self):
        """Tests LabTracks species is mapped to aind_data_schema species."""
        mouse = LabTracksSpecies.MOUSE.value

        subject_species = self.rh._map_species(mouse)
        none = self.rh._map_species("RANDOM_STUFF_dasfew32512")
        self.assertEqual(subject_species, Species.MUS_MUSCULUS)
        self.assertIsNone(none)
        self.assertIsNone(self.rh._map_species(None))

    def test_map_sex(self):
        """Tests that the LabTracks sex is mapped to aind_data_schema sex."""

        male = LabTracksSex.MALE.value
        female = LabTracksSex.FEMALE.value

        subject_male = self.rh._map_sex(male)
        subject_female = self.rh._map_sex(female)
        none = self.rh._map_sex("RANDOM_STUFF_afknewalofn1241413")

        self.assertEqual(subject_male, Sex.MALE)
        self.assertEqual(subject_female, Sex.FEMALE)
        self.assertIsNone(none)
        self.assertIsNone(self.rh._map_sex(None))

    def test_map_bg_strain(self):
        """Tests that the LabTracks strain is mapped correctly"""
        strain1 = LabTracksBgStrain.BALB_C.value
        strain2 = LabTracksBgStrain.C57BL_6J.value

        subject_strain1 = self.rh._map_to_background_strain(strain1)
        subject_strain2 = self.rh._map_to_background_strain(strain2)
        none = self.rh._map_to_background_strain("RANDOM_STUFF_afknewalofn1")

        self.assertEqual(subject_strain1, BackgroundStrain.BALB_c)
        self.assertEqual(subject_strain2, BackgroundStrain.C57BL_6J)
        self.assertIsNone(none)
        self.assertIsNone(self.rh._map_to_background_strain(None))


if __name__ == "__main__":
    unittest.main()



"""Tests NSB 2019 data models are created correctly"""

import json
import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import List, Tuple
from unittest import TestCase
from unittest import main as unittest_main

from aind_metadata_service.sharepoint.nsb2019.models import NSBList, Sex

if os.getenv("LOG_LEVEL"):  # pragma: no cover
    logging.basicConfig(level=os.getenv("LOG_LEVEL"))

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / ".." / ".."
SHAREPOINT_DIR = TEST_DIR / "resources" / "sharepoint" / "nsb2019" / "raw"
LIST_ITEM_FILE_NAMES = os.listdir(SHAREPOINT_DIR)
LIST_ITEM_FILE_PATHS = [SHAREPOINT_DIR / str(f) for f in LIST_ITEM_FILE_NAMES]


class TestNSB2019Models(TestCase):
    """Tests methods in NSBList2019 class"""

    @classmethod
    def setUpClass(cls):
        """Load json files before running tests."""
        cls.list_items = cls._load_json_files()

    @staticmethod
    def _load_json_files() -> List[Tuple[dict, str]]:
        """Reads test data into json"""
        list_items = []
        for file_path in LIST_ITEM_FILE_PATHS:
            with open(file_path) as f:
                contents = json.load(f)
            list_items.append((contents, file_path.name))
        list_items.sort(key=lambda x: x[1])
        return list_items

    def test_models_parsed(self):
        """Tests that the given list item files are parsed without error."""
        for index in range(len(self.list_items)):
            list_item = self.list_items[index][0]
            filename = self.list_items[index][1]
            logging.debug(f"Processing file: {filename}")
            nsb_model = NSBList.parse_obj(list_item)
            self.assertEqual(nsb_model.author_id, list_item.get("AuthorId"))

    def test_aberrant_data_parsed(self):
        """Tests that certain edge cases get handled correctly"""
        list_item = deepcopy(self.list_items[0][0])
        # Test that numeric entries instead of strings will also get parsed
        list_item["Age_x0020_at_x0020_Injection"] = "22"
        # Test that malformed or null types get mapped to None
        list_item["Inj1Type"] = "Wrong string"
        list_item["HPDurotomy"] = None
        # Check that the sex types are being mapped correctly
        list_item["Sex"] = "Female"
        nsb_model = NSBList.parse_obj(list_item)
        self.assertEqual("22", nsb_model.age_at_injection)
        self.assertIsNone(nsb_model.inj1_type)
        self.assertIsNone(nsb_model.hp_durotomy)
        self.assertEqual(Sex.FEMALE, nsb_model.sex)


if __name__ == "__main__":
    unittest_main()



"""Module to test SharePoint Client methods"""

import json
import logging
import os
import unittest
from pathlib import Path
from typing import List, Tuple
from unittest.mock import MagicMock, patch

from aind_metadata_service.sharepoint.client import SharePointClient

if os.getenv("LOG_LEVEL"):  # pragma: no cover
    logging.basicConfig(level=os.getenv("LOG_LEVEL"))

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / ".."
DIR_RAW_2019 = TEST_DIR / "resources" / "sharepoint" / "nsb2019" / "raw"
DIR_MAP_2019 = TEST_DIR / "resources" / "sharepoint" / "nsb2019" / "mapped"
LIST_ITEM_FILE_NAMES_2019 = os.listdir(DIR_RAW_2019)
sorted(LIST_ITEM_FILE_NAMES_2019)
LIST_ITEM_FILE_PATHS_2019 = [
    DIR_RAW_2019 / str(f) for f in LIST_ITEM_FILE_NAMES_2019
]
MAPPED_ITEM_FILE_NAMES_2019 = os.listdir(DIR_MAP_2019)
sorted(MAPPED_ITEM_FILE_NAMES_2019)
MAPPED_FILE_PATHS_2019 = [
    DIR_MAP_2019 / str(f) for f in MAPPED_ITEM_FILE_NAMES_2019
]
DIR_RAW_2023 = TEST_DIR / "resources" / "sharepoint" / "nsb2023" / "raw"
DIR_MAP_2023 = TEST_DIR / "resources" / "sharepoint" / "nsb2023" / "mapped"
LIST_ITEM_FILE_NAMES_2023 = os.listdir(DIR_RAW_2023)
sorted(LIST_ITEM_FILE_NAMES_2023)
LIST_ITEM_FILE_PATHS_2023 = [
    DIR_RAW_2023 / str(f) for f in LIST_ITEM_FILE_NAMES_2023
]
MAPPED_ITEM_FILE_NAMES_2023 = os.listdir(DIR_MAP_2023)
sorted(MAPPED_ITEM_FILE_NAMES_2023)
MAPPED_FILE_PATHS_2023 = [
    DIR_MAP_2023 / str(f) for f in MAPPED_ITEM_FILE_NAMES_2023
]


class TestSharepointClient(unittest.TestCase):
    """Class to test methods for SharePointClient."""

    @classmethod
    def setUpClass(cls):
        """Load json files before running tests."""
        cls.list_items_2019 = cls._load_json_files(year="2019")
        cls.list_items_2023 = cls._load_json_files(year="2023")

    @staticmethod
    def _load_json_files(year: str) -> List[Tuple[dict, dict, str]]:
        """Reads raw data and expected data into json"""
        list_items = []
        if year == "2019":
            list_item_file_paths = LIST_ITEM_FILE_PATHS_2019
        else:
            list_item_file_paths = LIST_ITEM_FILE_PATHS_2023
        for file_path in list_item_file_paths:
            mapped_file_path = (
                file_path.parent.parent
                / "mapped"
                / ("mapped_" + file_path.name)
            )
            with open(file_path) as f:
                contents = json.load(f)
            with open(mapped_file_path) as f:
                mapped_contents = json.load(f)
            list_items.append((contents, mapped_contents, file_path.name))
            list_items.sort(key=lambda x: x[2])
        return list_items

    @patch("aind_metadata_service.sharepoint.client.ClientContext")
    def test_empty_response(self, mock_sharepoint_client: MagicMock):
        """Tests that an empty response is generated if no data returned
        from NSB datatbases"""
        client = SharePointClient(
            nsb_site_url="some_url",
            nsb_list_title_2019="some_list_title2019",
            nsb_list_title_2023="some_list_title2023",
            client_id="some_client_id",
            client_secret="some_client_secret",
        )
        response = client.get_procedure_info(subject_id="12345")
        mock_sharepoint_client.assert_called()
        self.assertEqual(404, response.status_code)

    @patch("aind_metadata_service.sharepoint.client.ClientContext")
    def test_data_mapped(self, mock_sharepoint_client: MagicMock):
        """Tests that 200 response returned correctly."""

        inner_mock = MagicMock()
        mock_sharepoint_client.return_value.with_credentials.return_value = (
            inner_mock
        )
        mock_list_views = MagicMock()
        inner_mock.web.lists.get_by_title.return_value.views = mock_list_views
        mock_list_items = MagicMock()
        mock_list_views.get_by_title.return_value.get_items.return_value = (
            mock_list_items
        )
        list_item_2019_1 = self.list_items_2019[0][0]
        list_item_2023_1 = self.list_items_2023[0][0]
        mock_list_item2019 = MagicMock()
        mock_list_item2023 = MagicMock()
        mock_list_item2019.to_json.return_value = list_item_2019_1
        mock_list_item2023.to_json.return_value = list_item_2023_1
        mock_list_items.filter.side_effect = [
            [mock_list_item2019],
            [mock_list_item2023],
        ]

        client = SharePointClient(
            nsb_site_url="some_url",
            nsb_list_title_2019="some_list_title2019",
            nsb_list_title_2023="some_list_title2023",
            client_id="some_client_id",
            client_secret="some_client_secret",
        )

        expected_subject_procedures = []
        expected_subject_procedures.extend(self.list_items_2019[0][1])
        expected_subject_procedures.extend(self.list_items_2023[0][1])
        response = client.get_procedure_info(subject_id="12345")
        contents = json.loads(response.body.decode("utf-8"))
        expected_subject_procedures.sort(key=lambda x: str(x))
        contents["data"]["subject_procedures"].sort(key=lambda x: str(x))
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            expected_subject_procedures, contents["data"]["subject_procedures"]
        )


if __name__ == "__main__":
    unittest.main()
